#!/usr/bin/env pythonw

# NHigh - version 2.0.3 (eru_bahagon's BETA)
# Original file :
# http://metanet.2.forumer.com/index.php?showtopic=6697
# Written by jg9000  (jg9000 at gmail.com)
# Top 20 list contributed by thecalcgenius (PM me on the forums)
#
# You may freely modify and distribute this program, as long
# as I am credited.


'''Eru_bahagon has edited this programm.
Most edit are flaged. Visit the forum if you have any questions'''
# http://forum.ninjarobotyeti.com/viewtopic.php?f=20&t=9088

from nhighlib import *
from Tkinter import *
from tkhelper import *
from widgets import ToolTip
import tkFileDialog, tkMessageBox
import os, sys
import re

welcomeStr = '''\
NHigh version 2.0.3 - by jg9000
BETA by eru_bahagon

Select File -> Download to download the current highscores.
'''

WIN_TITLE = 'NHigh' #main window title
COLUMN_REP = 1 #if you want some kind of column representation (better layout)


######## ask dialog ########

class AskDlgField(object):
    def __init__(self, name, labelText=None, tooltip=None):
        self.name = name
        self.labelText = labelText
        self.tooltip = tooltip
        self.parentFrame = None #this will be set to the parent frame before build() is called
    def getResult(self):
        'Return the result, or throw AskDialog.ValidationFailedException'
        return 1
    def build(self):
        pass
    
    def buildEntry(self, var, default, width=None):
        if default is not None:
            var.set(default)
        self.entry = Entry(self.parentFrame, textvariable=var, width=width)
        self.initToolTip(self.entry)
        self.buildLabeledWidget(self.entry)

    def buildLabeledWidget(self, widget, sticky='ew', weight=0):
        label = self.parentFrame.addLabeledWidget(self.labelText, widget,
                                                  sticky=sticky, weight=weight)
        self.initToolTip(label)

    def initToolTip(self, widget):
        if self.tooltip:
            ToolTip(widget, self.tooltip)

    def readIntWidget(self, var, labelText, widget=None,
                      minValue=None, maxValue=None):
        try:
            val = int(var.get())
        except ValueError:
            msg = '%s: Please enter a number'%labelText
            raise AskDialog.ValidationFailedException(msg, widget)
        if (minValue is not None and val < minValue) or \
           (maxValue is not None and val > maxValue):
            if maxValue is not None and minValue is not None:
                cond = 'between %d and %d' % (minValue, maxValue)
            elif minValue is not None:
                cond = 'at least %d' % minValue
            else:
                cond = 'at most %d' % maxValue
            msg = '%s: must be %s' % (labelText, cond)
            raise AskDialog.ValidationFailedException(msg, widget)
        return val
        
class IntField(AskDlgField):
    def __init__(self, name, labelText, default=None, tooltip=None,
                 minValue=None, maxValue=None, width=None):
        AskDlgField.__init__(self, name, labelText, tooltip=tooltip)
        self.minValue = minValue
        self.maxValue = maxValue
        self.default = default
        self.width = width
    def build(self):
        self.var = StringVar()
        self.buildEntry(self.var, self.default, width=self.width)
    def getResult(self):
        return self.readIntWidget(self.var, self.labelText,
            self.entry, self.minValue, self.maxValue)

class TextField(AskDlgField):
    def __init__(self, name, labelText, default=None, tooltip=None,
                 allowEmpty=False):
        AskDlgField.__init__(self, name, labelText, tooltip=tooltip)
        self.allowEmpty = allowEmpty
        self.default = default
    def build(self):
        self.var = StringVar()
        self.buildEntry(self.var, self.default)
    def getResult(self):
        val = self.var.get()
        if not self.allowEmpty and not val:
            msg = '%s: Must not be empty' % self.labelText
            raise AskDialog.ValidationFailedException(msg, self.entry)
        return val

class MultiLineField(AskDlgField):
    def __init__(self, name, labelText, default=None, tooltip=None,
                 allowEmpty=False):
        AskDlgField.__init__(self, name, labelText, tooltip=tooltip)
        self.allowEmpty = allowEmpty
        self.default = default
    def build(self):
        frame = Frame(self.parentFrame)
        self.textWidget = Text(frame, wrap=CHAR, height=10)
        if self.default is not None:
            self.textWidget.insert(END, self.default)
        self.textWidget.grid(row=0, column=0, sticky='nesw')
        scroll = Scrollbar(frame, command=self.textWidget.yview)
        scroll.grid(row=0, column=1, sticky='ns')
        self.textWidget.config(yscrollcommand=scroll.set)
        frame.rowconfigure(0, weight=1)
        frame.columnconfigure(0, weight=1)
        
        self.initToolTip(self.textWidget)
        self.buildLabeledWidget(frame, sticky='nesw', weight=1)
    def getResult(self):
        val = self.textWidget.get('1.0',END)
        if not self.allowEmpty and not val:
            msg = '%s: Must not be empty' % self.labelText
            raise AskDialog.ValidationFailedException(msg, self.entry)
        return val

class BoolField(AskDlgField):
    def __init__(self, name, labelText, default, tooltip=None):
        AskDlgField.__init__(self, name, labelText, tooltip=tooltip)
        self.default = int(bool(default)) #convert to 1 or 0
    def build(self):
        self.var = IntVar()
        self.var.set(self.default)
        self.checkbox = Checkbutton(self.parentFrame,
            text=self.labelText, variable=self.var)
        self.initToolTip(self.checkbox)
        self.parentFrame.addWideWidget(self.checkbox, sticky='w')
    def getResult(self):
        return self.var.get()

class PlayerField(TextField):
    def __init__(self, name='player', labelText='Player Name', default=None,
                 tooltip=None, allowEmpty=False):
        if default is None:
            default = config.defaultPlayerName
        TextField.__init__(self, name, labelText, default=default,
                           tooltip=tooltip, allowEmpty=allowEmpty)
    def getResult(self):
        res = TextField.getResult(self)
        if re.search("['< ]", res):
            msg = '%s: Name contains invalid characters' % self.labelText
            raise AskDialog.ValidationFailedException(msg, self.entry)
        return res    

class FilenameField(TextField):
    def __init__(self, name='filename', labelText='File',
                 default=None, tooltip=None, mustExist=True):
        if default is None:
            default = config.defaultFilename
        TextField.__init__(self, name, labelText, default=default, tooltip=tooltip)
        self.mustExist = mustExist
        
    def build(self):
        self.var = StringVar()
        if self.default is not None:
            self.var.set(self.default)
        frame = Frame(self.parentFrame)
        self.entry = Entry(frame, textvariable=self.var)
        browseBtn = Button(frame, text='Browse...', command=self.onBrowse)
        self.entry.grid(column=0, row=0, sticky=E+W)
        browseBtn.grid(column=1, row=0, padx=1)
        frame.columnconfigure(0, weight=1)
        self.initToolTip(self.entry)
        self.buildLabeledWidget(frame)

    def onBrowse(self):
        fname = askFileName(self.parentFrame.getTopLevel(),
                            initialFile=self.var.get())
        if fname:
            self.var.set(fname)
            self.entry.update_idletasks()
            self.entry.selection_clear()
            self.entry.icursor(END)
            self.entry.xview(END)

    def getResult(self):
        val = TextField.getResult(self)
        if self.mustExist:
            fullval = val
            if '.' not in fullval:
                fullval += '.hs'
            if not os.path.exists(fullval):
                msg = '%s: File does not exist' % self.labelText
                raise AskDialog.ValidationFailedException(msg, self.entry)
        return val

#eru_bahagon added 3 following class
class RankField(IntField):
    def __init__(self, name='rank', labelText='Rank',
                 default=None, tooltip='Rank (0 to 19)'):
        IntField.__init__(self, name, labelText, default=default,
                          tooltip=tooltip, minValue=0, maxValue=19)

class IncludeEpField(BoolField):
    def __init__(self, name='inclEp', labelText='Include Episodes',
                default=1, tooltip='Tick to include Episodes'):
        BoolField.__init__(self, name, labelText,default=default, tooltip=tooltip)

class IncludeLvlField(BoolField):
    def __init__(self, name='inclLvl', labelText='Include Levels',
                default=1, tooltip='Tick to include Episodes'):
        BoolField.__init__(self, name, labelText, default= default, tooltip=tooltip)
#end of edit
    
class EpisodeNumField(IntField):
    def __init__(self, name='episode', labelText='Episode',
                 default=None, tooltip='Episode number (0 to 99)'):
        IntField.__init__(self, name, labelText, default=default,
                          tooltip=tooltip, minValue=0, maxValue=NUM_EPISODES-1)

class LevelNumField(AskDlgField):
    def __init__(self, name='level', labelText='',
                 default=None, tooltip='Level number (0 to 4)'):
        AskDlgField.__init__(self, name, labelText, tooltip=tooltip)
        self.default = default
        
    def build(self):
        self.entryVar = StringVar() #'' or '0' to '5'
        self.radioVar = IntVar() #0=levels, 1=episode

        if self.default == 5:
            self.radioVar.set(1)
        else:
            self.radioVar.set(0)
        if self.default in xrange(5):
            self.entryVar.set(str(self.default))
        frame = Frame(self.parentFrame)
        self.radioLevel = Radiobutton(frame, text='Level:',
            variable=self.radioVar, value=0, command=self.onRadio)
        self.radioEp = Radiobutton(frame, text='Episode',
            variable=self.radioVar, value=1, command=self.onRadio)
        self.entry = Entry(frame, width=3, textvariable=self.entryVar)
        if self.default==5:
            self.entry.config(state=DISABLED)
        self.radioLevel.grid(row=0,column=1, sticky=W)
        self.entry.grid(row=0, column=2, sticky=E+W)
        self.radioEp.grid(row=0,column=3)
        self.initToolTip(self.entry)
        self.parentFrame.addWideWidget(frame, sticky='w')

    def onRadio(self):
        if self.radioVar.get()==0:
            self.entry.config(state=NORMAL)
        else:
            self.entry.config(state=DISABLED)

    def getResult(self):
        if self.radioVar.get()==1:
            return 5 #episode
        else:
            return self.readIntWidget(self.entryVar, 'Level',
                                      self.entry, 0, 4)

class AskDlgResultObj(object):
    'Object used for results from AskDialog.run()'
    def _addFieldResult(self, fieldName, value):
        if fieldName=='#mergeResults':
            #merge members of result obj
            self.__dict__.update(value.__dict__)
        else:
            setattr(self, fieldName, value)
    
class AskDlgFrame(Frame):
    def __init__(self, master, fields):
        self.fields = fields
        self.hadRowWeight = False
        Frame.__init__(self, master)

    def build(self):        
        self.addFields()
        self.columnconfigure(1, weight=1)
        if not self.hadRowWeight:
            #add an empty row to keep everything at the top
            row = self.grid_size()[1]
            self.rowconfigure(row, weight=1)
        
    def addFields(self):
        for fld in self.fields:
            fld.parentFrame = self
            fld.build()

    def addLabeledWidget(self, labelText, widget, sticky='ew', weight=0):
        row = self.grid_size()[1]
        lbl = Label(self, text=labelText+':')
        lbl.grid(column=0, row=row, sticky=W+N)
        widget.grid(column=1, row=row, sticky=sticky)
        if weight:
            self.hadRowWeight = True
            self.rowconfigure(row, weight=weight)
        return lbl

    def addWideWidget(self, widget, sticky='ew', weight=0):
        row = self.grid_size()[1]
        widget.grid(row=row, columnspan=2, sticky=sticky)
        if weight:
            self.hadRowWeight = True
            self.rowconfigure(row, weight=weight)

    def getResult(self):
        res = AskDlgResultObj()
        for field in self.fields:
            fieldRes = field.getResult()
            res._addFieldResult(field.name, fieldRes)
        return res

    def getTopLevel(self):
        x = self
        while x and not hasattr(x, 'wm_iconify'):
            x = x.master
        return x

class AskDlgSubFrame(AskDlgField):
    def __init__(self, fields, name='#mergeResults'):
        AskDlgField.__init__(self, name)
        self.fields = fields
        
    def build(self):
        self.frame = AskDlgFrame(self.parentFrame, self.fields)
        self.frame.build()
        self.parentFrame.addWideWidget(self.frame)
        
    def getResult(self):
        return self.frame.getResult()

class AskDlgHorizSubFrame(AskDlgSubFrame):
    'a sub-frame that stores controls horizontally'
    def __init__(self, fields, name='#mergeResults'):
        AskDlgSubFrame.__init__(self, fields, name)

    def build(self):
        self.frame = AskDlgFrame(self.parentFrame, self.fields)
        self.frame.addLabeledWidget = self.addLabeledWidget
        self.frame.addWideWidget = self.addWideWidget
        self.frame.addFields()
        self.parentFrame.addWideWidget(self.frame, sticky='ew')

    def addLabeledWidget(self, labelText, widget, sticky='ew', weight=0):
        col = self.frame.grid_size()[0]
        lbl = Label(self.frame, text=labelText+':')
        lbl.grid(column=col, row=0)
        widget.grid(column=col+1, row=0, sticky=sticky, padx=3)
        self.frame.columnconfigure(col+1, weight=1)
        return lbl
        
    def addWideWidget(self, widget, sticky='ew', weight=0):
        col = self.frame.grid_size()[0]
        widget.grid(row=0, column=col, sticky=sticky)
        if weight:
            self.frame.columnconfigure(col, weight=weight)

class AskDlgSelectionSubFrame(AskDlgField):
    '''a sub-frame that chooses one of several fields to show.
       Only the shown widget will be in effect'''
    def __init__(self, subFields, subFieldLabels, subFieldNames=None,
                 name='#mergeResults', selectionName='selection'):
        AskDlgField.__init__(self, name)
        self.subFields = subFields
        if subFieldNames is None:
            subFieldNames = [str(x) for x in xrange(len(subFields))]
        self.subFieldNames = subFieldNames
        self.subFieldLabels = subFieldLabels
        self.selectionName = selectionName
        self.selectedFrame = None
        self.selectedFrameIdx = -1
        self.selectedFrameName = None
        
    def build(self):
        labelFrame = LabelFrame(self.parentFrame, text='Demo source')
        self.radioVar = StringVar()
        self.radioVar.set(self.subFieldNames[0])
        self.radios=[]
        for lbl,name in zip(self.subFieldLabels, self.subFieldNames):
            self.radios.append(Radiobutton(labelFrame, text=lbl,
                variable=self.radioVar, value=name, command=self.onRadio))
            self.radios[-1].grid(sticky='w')
            labelFrame.columnconfigure(0, weight=1)
        self.parentFrame.addWideWidget(labelFrame, sticky='ew')
        self.parentFrame.addWideWidget(Frame(self.parentFrame, height=15))
        
        self.masterFrame = Frame(self.parentFrame)
        self.frames = []
        for field in self.subFields:
            self.frames.append(AskDlgFrame(self.masterFrame, [field]))
            self.frames[-1].build()
        self.selectFrame(self.subFieldNames[0])
        self.masterFrame.columnconfigure(0, weight=1)
        self.masterFrame.rowconfigure(0, weight=1)
        self.parentFrame.addWideWidget(self.masterFrame, sticky='nesw', weight=1)

    def selectFrame(self, frameName):
        if frameName not in self.subFieldNames:
            return
        self.selectedFrameName = frameName
        self.selectedFrameIdx = self.subFieldNames.index(frameName)
        if self.selectedFrame:
            self.selectedFrame.grid_forget()
        self.selectedFrame = self.frames[self.selectedFrameIdx]
        self.selectedFrame.grid(row=0, column=0, sticky='nesw')
        self.masterFrame.columnconfigure(0, weight=1)
        self.masterFrame.rowconfigure(0, weight=1)

    def onRadio(self):
        self.selectFrame(self.radioVar.get())

    def getResult(self):
        res = self.selectedFrame.getResult()
        res._addFieldResult(self.selectionName, self.selectedFrameName)
        return res

class AskDialog(Dialog):
    def __init__(self, parent, title, description, fields):
        self.description = description
        self.fields = fields
        Dialog.__init__(self, parent, title, icon=app.iconfile)
        w = self.focus_get()
        if w and hasattr(w, 'selection_range'):
            w.selection_range(0, END)
        
    def body(self, master):
        descFrame = Frame(master)
        descFrame.grid(sticky=N+W+E)
        if self.description:
            master.rowconfigure(0, pad=5)
            descFrame.columnconfigure(0, weight=1)
            lbl = Label(descFrame, text=self.description, wraplength=400, justify=LEFT)
            lbl.grid(row=0,column=0, sticky=W)
            line = Frame(descFrame, borderwidth=1, relief=SUNKEN)
            line.grid(row=1, sticky=E+W, pady=5)
            Frame(line).grid()
        self.fieldFrame = AskDlgFrame(master, self.fields)
        self.fieldFrame.grid(sticky=N+S+E+W)
        self.fieldFrame.build()
        master.columnconfigure(0, weight=1)
        master.rowconfigure(1, weight=1)

    class ValidationFailedException(Exception):
        def __init__(self, msg, widget=None):
            Exception.__init__(self, msg)
            self.widget = widget

    def validate(self):
        try:
            self.result = self.fieldFrame.getResult()
            return 1
        except AskDialog.ValidationFailedException as e:
            tkMessageBox.showerror('Error',str(e), parent=self)
            if e.widget:
                e.widget.focus_set()
                if hasattr(e.widget, 'selection_range'):
                    e.widget.selection_range(0, END)
            return 0

######## config dlg ########

class PlayerListField(AskDlgField):
    def __init__(self, name, labelText, addDlgTitle, addDlgDesc,
                 default=None, tooltip=None, optionsDlg=None):
        AskDlgField.__init__(self, name, labelText, tooltip=tooltip)
        self.default = default or []
        self.optionsDlg = optionsDlg
        self.addDlgTitle = addDlgTitle
        self.addDlgDesc = addDlgDesc
    def build(self):
        frame = Frame(self.parentFrame)
        lbl = Label(frame, text=self.labelText+':')
        lbl.grid(column=0, row=0, sticky=W, columnspan=2)
        self.list = Listbox(frame, selectmode=EXTENDED)
        self.list.grid(column=0, row=1, sticky=N+S+E+W)
        for x in self.default:
            self.list.insert(END, x)
        scrollbar = Scrollbar(frame, orient=VERTICAL, command=self.list.yview)
        self.list.config(yscrollcommand=scrollbar.set)
        scrollbar.grid(column=1, row=1, sticky=N+S)

        #empty column for spacing
        Frame(frame).grid(row=1,column=2,padx=5)

        buttonFrame = Frame(frame)
        buttonFrame.grid(column=3, row=1, sticky=N+E)
        btnAdd = Button(buttonFrame, text='Add...', command=self.addPlayer)
        btnAdd.grid(row=0, sticky=E+W, pady=3, ipadx=5)
        btnRemove = Button(buttonFrame, text='Remove', command=self.removePlayer)
        btnRemove.grid(row=1, sticky=E+W, pady=3, ipadx=5)

        frame.rowconfigure(1, weight=4)
        frame.columnconfigure(0, weight=1)
        
        self.parentFrame.addWideWidget(frame, sticky='nsew', weight=1)

    def addPlayer(self):
        fields = [PlayerField(default='')]
        dlg = AskDialog(parent=self.parentFrame.getTopLevel(),
            title=self.addDlgTitle, description=self.addDlgDesc, 
            fields=fields)
        res = dlg.run()
        if res:
            self.list.insert(END, res.player)

    def removePlayer(self):
        sel = list(self.list.curselection())
        if not sel: return
        sortList(sel, reverse = True)
        for x in sel:
            self.list.delete(x)

    def getResult(self):
        return list(self.list.get(0,END))
        
class OptionsDlg(AskDialog):
    def __init__(self, parent):
        fields = [
            PlayerField(name='defaultPlayerName', labelText='Default Player Name',
                        default=config.defaultPlayerName, allowEmpty=True),
            TextField(name='defaultFilename', labelText='Default File name',
                      default=config.defaultFilename, allowEmpty=True),
            BoolField(name='allowZerothTies', labelText='Allow 0th ties',
                      default=config.allowZerothTies, tooltip=
                      'If selected: Count all tied 0th in a level (default)\n'
                      'If not selected: Only first 0th counts'),
            PlayerListField(name='ignoredPlayers', labelText='Players to ignore',
                            addDlgTitle = 'Add Ignored Player',
                            addDlgDesc = 'Enter the name of a player which will be ignored',
                            default=config.ignoredPlayers)
        ]
        
        AskDialog.__init__(self, parent, 'NHigh Options', description=None, fields=fields)
        
    def run(self):
        res = AskDialog.run(self)
        if res:
            config.defaultFilename = res.defaultFilename
            config.ignoredPlayers = res.ignoredPlayers
            config.defaultPlayerName = res.defaultPlayerName
            config.allowZerothTies = res.allowZerothTies
            try:
                config.save()
            except (IOError, OSError):
                tkMessageBox.showwarning('Error','Error saving configuration to file:\n%s'%config.configFile)
        return res
            
##############

def getPlural(num):
    if num==1:
        return ''
    else:
        return 's'

def getOrdinal(num):
    if (num%100)/10==1:
        return '%dth'%num
    return '%d%s' % (num, {1:'st',2:'nd',3:'rd'}.get(num,'th'))

def askFileName(parent, title=None, save=False, initialFile=None):
    fileTypes = (
        ('Highscore files','*.hs'),
        ('All files','*')
        )
    if not title:
        if save:
            title = 'Save As'
        else:
            title = 'Open'
#apparently, something weird happens on mac here
    platform = getPlatform()
    if platform == 'darwin' :
        defExt = ''
    else : 
        defExt = '.hs'
    if save:
        func = tkFileDialog.asksaveasfilename
    else:
        func = tkFileDialog.askopenfilename
    fileName = func(parent=parent,
        filetypes = fileTypes,
        defaultextension = defExt,
        title = title,
        initialfile = initialFile)
    if fileName:
        fileName = fileName.replace('/', os.sep)
    return fileName


class SubMenu(object):
    def __init__(self, label, items):
        self.label = label
        self.items = items
class MenuItem(object):
    def __init__(self, label, command, key=None):
        self.label = label
        self.command = command
        self.wrappedCommand = app.eventWrapper(command)
        self.key = key
        
class ContextMenuItem(MenuItem):
    'This is like MenuItem, except it sets members according to'
    'which player/level/etc. the context menu was shown for.'
    def __init__(self, label, command, key=None):
        def wrappedCommand():
            if app.lastRightClickedPlayer :
                app.selectedPlayer = app.lastRightClickedPlayer.lstrip()
            else : 
                app.selectedPlayer = app.lastRightClickedPlayer
            app.selectedLevel = app.lastRightClickedLevel
            try:
                command()
            finally:
                app.selectedPlayer = None
                app.selectedLevel = None

        newcommand = command            
        if newcommand:
            newcommand = wrappedCommand
        MenuItem.__init__(self, label, newcommand, key)
    
class MenuSeparator(object):
    pass

######## app ########
        
class NHighApp(object):
    def __init__(self, root):
        self.root = root
        self.curHS = None
        self.selectedPlayer = None
        self.lastRightClickedPlayer = None
        self.selectedLevel = None
        self.lastRightClickedLevel = None
#eru_bahagon added next obj
#it is a dictionary with player as key and their corresponding table of all online scores
        self.onlineScores = None

    def build(self):        
        self.buildRootWin(self.root)

    def eventWrapper(self, func):
        '''This wraps around event handlers to provide exception
        handling.'''
        def wrapper(event=None):
            try:
                func()
            except (IOError, OSError, NHighError) as e:
                tkMessageBox.showwarning('Error', str(e), parent=self.root)
        return wrapper
                    
    def checkCurHS(self):
        if not self.curHS:
            raise NHighError('''\
No highscores loaded.

Download the current highscores using File->Download,
or load a previously saved file.''')
        
    def setCurHS(self, hs):
        self.curHS = hs
        self.txtCurHSDate.set(str(hs.timestamp))
        
    def downloadHS(self):
        def onProgress(progress):
            self.root.title(WIN_TITLE+' - %d%%'%int(progress*100))
        downloader = HSDownloader()
        dlg = MeterDialog(self.root,'Downloading',downloader,
                          useThread=True, onProgress=onProgress,
                          icon=self.iconfile)
        if dlg.run():
            self.setCurHS(downloader.result())
        self.root.title(WIN_TITLE)

    def loadHS(self):
        fname = askFileName(self.root, initialFile=config.defaultFilename)
        if fname:
            self.setCurHS( loadScores(fname) )
    
    def autoLoadHS(self):
        try :
            initialdir = os.getcwd()
            if getPlatform() == 'darwin' :
                fileName = initialdir + '/Documents/nhigh/' + config.defaultFilename
            else :
                fileName = initialdir + '/' + config.defaultFilename
            fileName = fileName.replace('/', os.sep)
            self.setCurHS( loadScores(fileName) )
        except : pass

    def saveHS(self):
        self.checkCurHS()
        fname = askFileName(self.root, save=True, initialFile=config.defaultFilename)
        if fname:
            saveScores(self.curHS, fname)

#eru added following function
    def downloadOnlineScores(self, players):
        def onProgress(progress):
            self.root.title(WIN_TITLE+' - %d%%'%int(progress*100))
        
        downloader = OnlineScoresDownloader(players)
        dlg = MeterDialog(self.root,'Downloading',downloader,
                          useThread=True, onProgress=onProgress,
                          icon=self.iconfile)
        if dlg.run():
            res = downloader.result()
            if not self.onlineScores :
                self.onlineScores = res
            else : 
                for k in res.keys():
                    self.onlineScores[k] = res[k]
                    
        self.root.title(WIN_TITLE)
#end of edit

    def askAndLoadOldScores(self, title, desc):
        fields = [FilenameField(labelText='Old highscore file',
                                tooltip='A previously saved highscore file\nto compare with the current one')]
        dlg = AskDialog(self.root, title, desc, fields)
        res = dlg.run()
        if res:
            return loadScores(res.filename)

    def askOldScoresAndPlayer(self, title, desc):        
        fields = [FilenameField(labelText = 'Old highscore file', tooltip='A previously saved highscore file\nto compare with the current one'),
                  PlayerField(default=self.selectedPlayer)]
        dlg = AskDialog(self.root, title, desc, fields)
        res = dlg.run()
        if not res:
            return (None, None)
        return (loadScores(res.filename), res.player)

    def askPlayer(self, title, desc):
        if self.selectedPlayer:
            return self.selectedPlayer
        fields = [PlayerField()]
        dlg = AskDialog(self.root, title, desc, fields)
        res = dlg.run()
        if not res:
            return None
        return res.player

    def askLevelOrEpisode(self, title, desc):
        if self.selectedLevel:
            return self.selectedLevel
        fields = [EpisodeNumField(), LevelNumField()]
        dlg = AskDialog(self.root, title, desc, fields)
        res = dlg.run()
        if not res:
            return (None,None)
        else:
            return (res.episode, res.level)

#eru_bahagon added 4 following methods
    def askNumber(self, title, desc):
        fields = [IntField(name='nbr', labelText = 'Number', minValue = 2)]
        dlg = AskDialog(self.root, title, desc, fields)
        res = dlg.run()
        if not res : 
            return None
        return res.nbr
    
    def askRanks(self, title, desc):
        fields = [RankField(name='r1', labelText='Rank 1'), RankField(name='r2', labelText='Rank 2'), 
                    IncludeEpField(), IncludeLvlField()]
        dlg = AskDialog(self.root, title, desc, fields)
        res = dlg.run()
        if not res : 
            return (None, None, None, None)
        return (res.r1, res.r2, res.inclEp, res.inclLvl)
    
    def askPlayerandRanks(self, title, desc):
        if self.selectedPlayer :
            fields = [RankField(name='r1', labelText='Rank 1'), RankField(name='r2', labelText='Rank 2')]
            dlg = AskDialog(self.root, title, desc, fields)
            res = dlg.run()
            if not res : 
                return (None, None, None)
            return (self.selectedPlayer, res.r1, res.r2)
        else :
            fields = [PlayerField(), RankField(name='r1', labelText='Rank 1'), RankField(name='r2', labelText='Rank 2')]
            dlg = AskDialog(self.root, title, desc, fields)
            res = dlg.run()
            if not res : 
                return (None, None, None)
            return (res.player, res.r1, res.r2)
    
    def askPlayers(self, title, desc):
        if self.selectedPlayer :
            fields = [PlayerField(labelText='Player 2'), IncludeEpField(default=0)]
            dlg = AskDialog(self.root, title, desc, fields)
            res = dlg.run()
            if not res :
                return (None, None, None)
            return (self.selectedPlayer, res.player, res.inclEp)
        else :
            fields = [PlayerField(name='p1', labelText='Player 1'), PlayerField(name='p2', labelText='Player 2'),
                IncludeEpField(default=0)]
            dlg = AskDialog(self.root, title, desc, fields)
            res = dlg.run()
            if not res :
                return (None, None)
            return (res.p1, res.p2)
            
#end of edit
        
    def addPlayerToIgnoreList(self):
        player = self.selectedPlayer
        if not player:
            return
        if player.lower() in config.ignoredPlayersSet:
            raise NHighError('This player is already in the ignore list')

        if not tkMessageBox.askyesno('Ignore Player?', '''\
Are you sure you want to add %s to the ignore list?
This will cause the player to be ignored when showing
high scores and general ranking.''' % player, default=tkMessageBox.NO,
                                     parent=self.root):
            return
        config.ignoredPlayers.append(player)
        try:
            config.save()
        except (IOError, OSError):
            raise NHighError('Error saving configuration to file:\n%s'%config.configFile)

    def player0thChangesReport(self):
        self.checkCurHS()
        oldHS, player = self.askOldScoresAndPlayer('Player 0th Changes Report',
            'This shows on which levels the player gained or lost 0th scores. The currently loaded highscores are compared with a previously saved file.')
        if not oldHS: return

        self.setText('Player 0th changes report for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        cur0th = getPlayerZeroth(self.curHS, player)
        old0th = getPlayerZeroth(oldHS, player)
        cur0thDict = dict([((ep,lvl),entry) for ep,lvl,entry in cur0th])
        old0thDict = dict([((ep,lvl),entry) for ep,lvl,entry in old0th])
        gained = []; lost = []; improved = []; tied = []
        for ep,lvl,curEntry in cur0th:
            if (ep,lvl) in old0thDict:
                oldEntry = old0thDict[(ep,lvl)]
                if curEntry.score != oldEntry.score:
                    improved.append((ep,lvl,oldEntry,curEntry))
            else:
                oldEntry = oldHS.table[ep][lvl][0]
                if curEntry != self.curHS.table[ep][lvl][0]:
                    tied.append((ep,lvl,oldEntry, curEntry))
                else:
                    gained.append((ep,lvl,oldEntry, curEntry))
        for ep,lvl,oldEntry in old0th:
            if not (ep,lvl) in cur0thDict:
                lost.append((ep,lvl,oldEntry, self.curHS.table[ep][lvl][0]))

        self.showPlayerChanges(lost, 'Lost')
        self.showPlayerChanges(gained, 'Gained')
        self.showPlayerChanges(tied, 'Tied')
        self.showPlayerChanges(improved, 'Improved')
        oldScores = getPlayerScores(oldHS, player)
        curScores = getPlayerScores(self.curHS, player)
        self.addLine('\nBefore: %d 0th, %d scores.  After: %d 0th, %d scores' %
              (len(old0th),len(oldScores),len(cur0th),len(curScores)))

    def showPlayerChanges(self, changes, didwhat):
        if not changes: return
        self.addLine('%s %d score%s:'%(didwhat, len(changes),
                           getPlural(len(changes))), 'h2')
        for ep,lvl,oldentry,newentry in changes:
            self.addLine('   %s - Old: %s  New: %s' %
                (self.dispLevel(ep,lvl),self.dispEntry(oldentry),
                 self.dispEntry(newentry)))

    def newZerothReport(self):
        self.checkCurHS()
        oldHS = self.askAndLoadOldScores('New 0th Scores',
            'This report shows new 0th scores by all players. The currently loaded highscores are compared with a previously saved file (selected below).')
        if not oldHS:
            return
        imps = findZerothImprovements(oldHS, self.curHS)
        self.setText('New 0th Report\n\n','h1')
        for ep,lvl,oldentry,newentry in imps:
            self.addLine('%s - Old: %s  New: %s' %
                (self.dispLevel(ep,lvl),self.dispEntry(oldentry),
                 self.dispEntry(newentry)))
        self.addLine('\nTotal: %d changes' % len(imps))

    def player20ChangesReport(self):
        self.checkCurHS()
        oldHS, player = self.askOldScoresAndPlayer('Player Top-20 Changes Report',
            'This shows on which levels the player gained or lost top-20 scores. The currently loaded highscores are compared with a previously saved file.')
        if not oldHS: return
        self.setText('Player top-20 changes report for %s:\n\n' %
                     self.dispPlayer(player), 'h1')

        cur20 = getPlayerScores(self.curHS, player)
        old20 = getPlayerScores(oldHS, player)
        cur20Dict = dict([((ep,lvl),(rank,entry))
                          for ep,lvl,rank,entry in cur20])
        old20Dict = dict([((ep,lvl),(rank,entry))
                          for ep,lvl,rank,entry in old20])
        cur20Set = set(cur20Dict.iterkeys())
        old20Set = set(old20Dict.iterkeys())
        gained = []; lost = []
        for ep,lvl in cur20Set-old20Set:
            rank,entry = cur20Dict[(ep,lvl)]
            gained.append((ep,lvl,rank, entry))
        for ep,lvl in old20Set-cur20Set:
            rank,entry = old20Dict[(ep,lvl)]
            lost.append((ep,lvl,rank, entry))

        self.showPlayer20Changes(gained, 'Gained')
        self.showPlayer20Changes(lost, 'Lost')
        self.addLine('\nBefore: %d scores.  After: %d scores' %
              (len(old20),len(cur20)))

    def showPlayer20Changes(self, changes, didwhat):
        if not changes: return
        self.addLine('%s %d score%s:'%(didwhat, len(changes),
                                       getPlural(len(changes))), 'h2')
        for ep,lvl,rank,entry in changes:
            self.addLine('   %s - %4s - %s' %
                (self.dispLevel(ep,lvl), getOrdinal(rank),
                 self.dispTime(entry.time)))

    def playerStatisticsReport(self):        
        self.checkCurHS()
        player = self.askPlayer('Player Statistics Report',
            'This report shows the number of top-20 and 0th scores for the player')
        if not player: return
        
        stats = getAllPlayerStats(self.curHS)
        plyrstats = stats.get(player.lower(), [player,0,0,0])
        more_0th = len([x for x in stats.values() if x[1]>plyrstats[1]])
        more_scores = len([x for x in stats.values() if x[2]>plyrstats[2]])
        self.setText( 'Statistics for %s:\n\n' % self.dispPlayer(plyrstats[0]), 'h1')
        if plyrstats[1]>0:
            self.addLine('Number of 0th: %d  (ranked #%d)'%(plyrstats[1] ,more_0th))
        else:
            self.addLine('Number of 0th: %d'%plyrstats[1])
        if plyrstats[2]>0:
            self.addLine('Number of top-20 highscores: %d  (ranked #%d)'%(plyrstats[2] ,more_scores))
        else:
            self.addLine('Number of top-20 highscores: %d'%plyrstats[2])

    def player0thList(self):
        self.checkCurHS()
        player = self.askPlayer('Player 0th List',
            'Displays a list of all 0th scores held by a player')
        if not player: return

        self.setText('List of 0th scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        lst = getPlayerZeroth(self.curHS, player)
        for ep_num, lvl_num, entry in lst:
            self.addLine('%s - %s' % (self.dispLevel(ep_num, lvl_num),
                                         self.dispTime(entry.time)))
        self.addLine('\nTotal: %d 0th scores' % len(lst))

    def player20List(self):
        self.checkCurHS()
        player = self.askPlayer('Player Top-20 List',
            'Displays a list of all top-20 scores held by a player')
        if not player: return

        self.setText('List of top-20 scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        twenty = getPlayerScores(self.curHS, player)
        for i, (ep_num, lvl_num, r, entry) in enumerate(twenty):
            self.addLine('%s - %4s - %s' %
                (self.dispLevel(ep_num, lvl_num), getOrdinal(r),
                 self.dispTime(entry.time)))
        self.addLine('\nTotal: %d top-20 scores' % len(twenty))
        
#eru_bahagon added the 7 following function
    def playerEp20List(self):
        self.checkCurHS()
        player = self.askPlayer('Player Top-20 List',
            'Displays a list of all episode top-20 scores held by a player')
        if not player: return

        self.setText('List of episode top-20 scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        twenty = getPlayerEpScores(self.curHS, player)
        for i, (ep_num, r, entry) in enumerate(twenty):
            self.addLine('%s - %4s - %s' %
                (self.dispLevel(ep_num, 5), getOrdinal(r),
                 self.dispTime(entry.time)))
        self.addLine('\nTotal: %d episode top-20 scores' % len(twenty))
    
    def playerWorstLvlList(self):
        self.checkCurHS()
        player = self.askPlayer('Player Worst Level Scores List',
            'Displays a list of the worst scores held by a player (in terms of diff to 0th)')
        if not player : return
        
        self.setText('List of worst level top-20 scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1') 
        list = getPlayerWorstLvlScores(self.curHS, player)
        for i, (ep, lvl, rank, entry, diff) in enumerate(list[:100]) :
            self.addLine('%2s - %s - %4s - %s Diff: %s' %
                (i, self.dispLevel(ep, lvl), getOrdinal(rank), 
                self.dispTime(entry.time), self.dispTime(diff)))
    
    def playerWorstEpList(self):
        self.checkCurHS()
        player = self.askPlayer('Player Worst Episode Scores List',
            'Displays a list of the worst episode scores held by a player (in terms of diff to 0th)')
        if not player : return
        
        self.setText('List of worst episode top-20 scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1') 
        list = getPlayerWorstEpScores(self.curHS, player)
        for i, (ep, rank, entry, diff) in enumerate(list[:40]) :
            self.addLine('%2s - %s - %4s - %s Diff: %s' %
                (str(i), self.dispLevel(ep, 5), getOrdinal(rank), 
                self.dispTime(entry.time), self.dispTime(diff)))
        
        
    def _doPlayerXYList(self, title, player, rank1 = 0, rank2 = 9):
        #list of all the scores of a player (between two ranks)
        self.checkCurHS()
        if not player: 
            return

        self.setText(title+'\n\n', 'h1')
        
        twenty = getPlayerScores(self.curHS, player)
        len = 0
        for i, (ep_num, lvl_num, r, entry) in enumerate(twenty):
            if r <= rank2 and r >= rank1 :
                len += 1
                self.addLine('%s - %4s - %s' %
                    (self.dispLevel(ep_num, lvl_num), getOrdinal(r),
                    self.dispTime(entry.time)))
        self.addLine('\nTotal: %d scores' % len)
    
    
    def playerXYList(self):
        self.checkCurHS()
        player, r1, r2 = self.askPlayerandRanks('Player Top-X Scores', 
            "Shows player's scores between the specified ranks (both included, rank1 <= rank2)")
        if r1 == None or r2 == None or not player:
            return
        
        if r1 > r2 :
            self.setText('I told you that rank1 must be smaller or equal to rank2.')
            return
        
        if r1 == 1 and r1 == r2:
            self._doPlayer1stList(player)
            return
        
        if r1 == 0 and r2 == 0:
            str = 'List of 0th scores (no tie)'
        elif r1 == 0:
            str = 'List of Top-%d scores' % (int(r2)+1)
        elif r1 == r2 : 
            str = 'List of %s scores' % getOrdinal(int(r1))
        else :
            str = 'List of %s-%s scores' % (getOrdinal(int(r1)), getOrdinal(int(r2))) 
        
        str += " for %s:" % self.dispPlayer(player)
        self._doPlayerXYList(str, player, rank1 = r1, rank2 = r2)
        
        
    def playerNon20List(self):
        #list of all non-top 20 scores for a given player
        self.checkCurHS()
        player = self.askPlayer('Player Non-Top-20 List',
            'Displays a list of all non-top-20 scores held by a player')
        if not player: return
        
        nontwenty = getPlayerNonScores(self.curHS, player)
        if len(nontwenty) == 0:
            self.setText('No non-top-20 score for %s' % player)
        else :
            self.setText('List of non-top-20 scores for %s:\n\n' %
                         self.dispPlayer(player), 'h1')
            for (ep_num, lvl_num) in nontwenty:
                self.addLine('%s' %
                    self.dispLevel(ep_num, lvl_num))
            self.addLine('\nTotal: %d non-top-20 scores' % len(nontwenty))
            
    def _doPlayer1stList(self, player):
        #list the players who have a while while 'player' is 1st
        self.checkCurHS()
        #player = self.askPlayer('Player 1st List',
        #    'Displays a list of all 1st held by a player, indicating who has the 0th')
        if not player : return
        
        self.setText('List of 1st for %s:\n\n' % self.dispPlayer(player), 'h1')
        first = getPlayer1st(self.curHS, player)
        len = 0
        for zeroth in first.keys():
            count = 0
            self.addLine('%s as 0th holder :' % self.dispPlayer(zeroth))
            for ep_num, lvl_num, entry in first[zeroth]:
                count += 1
                len += 1
                self.addLine('%s - %s' % (self.dispLevel(ep_num, lvl_num), 
                    self.dispTime(entry.time)))
            self.addLine('Total for %s : %d\n\n' % (self.dispPlayer(zeroth), count))
        self.addLine('\nTotal: %d scores' % len)
                
    
    def playerComparison(self):
        #compare scores of two players
        self.checkCurHS()
        players = self.askPlayers('Player Comparison', 
            'Displays a list of levels where Player 2 is faster than Player 1')
        if not players : return
        
        p1 = players[0]
        p2 = players[1]
                
        self.setText('List of better runs for %s over %s :\n\n' 
            % (self.dispPlayer(p2), self.dispPlayer(p1)), 'h1')
        lst = getBetterRuns(self.curHS, p1, p2)
        diffTot = 0
        for i, (ep, lvl, diff) in enumerate(lst):
            diffTot += diff
            self.addLine('%3s - %s   Diff: %s' % (i, self.dispLevel(ep,lvl), self.dispTime(diff)))
        self.addLine('\nTotal Difference: %s' % self.dispTime(diffTot))
        self.addLine('\nTotal: %d scores' % len(lst))
        
#end of edit

    def difficultLevelsReport(self, rank1, rank2):
        self.checkCurHS()
        self.setText('Levels with highest difference between %s and %s scores:\n\n' %
              (getOrdinal(rank1), getOrdinal(rank2)), 'h1')
        lst = getLevelRankDifferences(self.curHS, rank1, rank2)
        for i, (ep, lvl, diff, entry1, entry2) in enumerate(lst[:50]):
            self.addLine('%2s - %s - %s  Diff: %s' % (i, self.dispLevel(ep,lvl),
                         self.dispEntry(self.curHS.table[ep][lvl][0]),
                         self.dispTime(diff)))
    
#eru_bahagon edit (added the 5 following functions)
    def difficultEpsReport(self, rank1, rank2):
        self.checkCurHS()
        self.setText('Episodes with highest difference between %s and %s scores:\n\n' %
              (getOrdinal(rank1), getOrdinal(rank2)), 'h1')
        lst = getEpRankDifferences(self.curHS, rank1, rank2)
        for i, (ep, diff, entry1, entry2) in enumerate(lst[:30]):
            self.addLine('%2s - %s - %s  Diff: %s' % (i, self.dispLevel(ep,5),
                         self.dispEntry(self.curHS.table[ep][5][0]),
                         self.dispTime(diff)))
                         
    def easyLevelsReport(self, rank1, rank2, incl_tie=True):
        self.checkCurHS()
        self.setText('Levels with smallest difference between %s and %s scores:\n\n' %
              (getOrdinal(rank1), getOrdinal(rank2)), 'h1')
        lst = getLevelRankDifferences(self.curHS, rank1, rank2, rev=False, tie=incl_tie)
        for i, (ep, lvl, diff, entry1, entry2) in enumerate(lst[:100]):
            self.addLine('%2s - %s - %s  Diff: %s' % (i, self.dispLevel(ep,lvl),
                         self.dispEntry(self.curHS.table[ep][lvl][0]),
                         self.dispTime(diff)))
                         
    def easyEpsReport(self, rank1, rank2):
        self.checkCurHS()
        self.setText('Episodes with smallest difference between %s and %s scores:\n\n' %
              (getOrdinal(rank1), getOrdinal(rank2)), 'h1')
        lst = getEpRankDifferences(self.curHS, rank1, rank2, rev=False)
        for i, (ep, diff, entry1, entry2) in enumerate(lst[:30]):
            self.addLine('%2s - %s - %s  Diff: %s' % (i, self.dispLevel(ep,5),
                         self.dispEntry(self.curHS.table[ep][5][0]),
                         self.dispTime(diff)))
						 
    def mostImprovableEps(self):
        self.checkCurHS()
        self.setText('Episodes with largest difference between their level 0th and ep 0th:\n\n', 'h1')
        lst = getEpLevelDifference(self.curHS)
        for i, (ep, diff, entry) in enumerate(lst[:50]):
            self.addLine('%2s - %s - %s Diff : %s' % (i, self.dispLevel(ep, 5), 
                        self.dispEntry(self.curHS.table[ep][5][0]), self.dispTime(diff)))
	
    def leastImprovableEps(self):
        self.checkCurHS()
        self.setText('Episodes with smallest difference between their level 0th and ep 0th:\n\n', 'h1')
        lst = getEpLevelDifference(self.curHS, rev=False)
        for i, (ep, diff, entry) in enumerate(lst[:50]):
            self.addLine('%2s - %s - %s Diff : %s' % (i, self.dispLevel(ep, 5), 
                        self.dispEntry(self.curHS.table[ep][5][0]), self.dispTime(diff)))
    
    def ownedEp(self):
        self.checkCurHS()
        self.setText('Episode Owned (or partially):\n', 'h1')
        dic = getOwnedEp(self.curHS)
        lst = dic.keys()
        sortList(lst, reverse=True)
        for num in lst:
            if num == 6:
                self.addLine('\nEpisode Owned:')
            else : 
                self.addLine('\nPlayers with %d 0th in an episode:' % num)
            for ep, player in dic[num]:
                self.addLine('%s- %s' % (self.dispLevel(ep, 5), self.dispPlayer(player)))
            

        
#end of edit
	
    def mostZerothReport(self):
        self._doScoresReport(1, '0th Rankings')

    def mostScoresReport(self):
        self._doScoresReport(2, 'Top-20 Rankings')

    def mostZerothReportNoEpisodes(self):
        self._doScoresReport(1, 'Level 0th Rankings',
                             includeEpisodes=False)


#eru_bahagon edit (add parameter includeLevels and two ranks)
    def _doScoresReport(self, itemNumber, title, numScores=100,
                        includeEpisodes=True, includeLevels=True, rank1=0, rank2=9):
        self.checkCurHS()
        self.setText(title+'\n\n', 'h1')
        stats = getAllPlayerStats(self.curHS, includeEpisodes, includeLevels, rank1, rank2).values()
        stats = [(x[itemNumber],x[0]) for x in stats if x[itemNumber]>0] #x iterates over players (keys)
        sortList(stats, reverse=True)
        numScores = min(numScores, len(stats))
        while numScores<len(stats) and stats[numScores-1][0]==stats[numScores][0]:
            numScores += 1
        for i in xrange(numScores):
            shownIndex = i
            if i>0 and stats[i][0] == stats[i-1][0]:
                shownIndex = '-'
            self.addLine('%2s  %s - %d'%(shownIndex,
                self.dispPlayer(stats[i][1]), stats[i][0]))
	
#eru_bahagon edit (added following 17 functions)
    def mostZerothReportNoLevels(self):
        self._doScoresReport(1, 'Episode 0th Rankings', includeLevels=False)
    
    def mostEpScoresReport(self):
        self._doScoresReport(2, 'Episode Top-20 Rankings', includeLevels=False)
        
    def mostLvlScoresReport(self):
        self._doScoresReport(2, 'Level Top-20 Rankings', includeEpisodes=False)
    
    def mostXYReport(self):
        self.checkCurHS()
        r1, r2, inclEp, inclLvl = self.askRanks('Top-X Rankings', 
            'Show the rankings between the specified ranks (both included, rank1 <= rank2)')
        if r1 == None or r2 == None :
            return
        if not inclEp and not inclLvl :
            self.setText('Yeah, you should include level or episode.')
            return
        if r1 > r2 :
            self.setText('I told you that rank1 must be smaller or equal to rank2.')
            return
        
        if r1 == 0 and r2 == 0:
            str = '0th Rankings (no tie)'
        elif r1 == 0:
            str = 'Top-%d Rankings' % (int(r2)+1)
        elif r1 == r2 : 
            str = '%s Rankings' % getOrdinal(int(r1))
        else :
            str = '%s-%s Rankings' % (getOrdinal(int(r1)), getOrdinal(int(r2))) 
        
        if not inclEp :
            str += ' (without Episodes)'
        if not inclLvl :
            str += ' (without Levels)'
        self._doScoresReport(3, str, includeEpisodes=inclEp, includeLevels=inclLvl, rank1 = r1, rank2 = r2)
        
                             
    def mostPoints(self):
        self._doPointsRankings('Total Points Rankings')
    
    def mostLvlPoints(self):
        self._doPointsRankings('Level Points Rankings', includeEpisodes=False)
    
    def mostEpPoints(self):
        self._doPointsRankings('Episode Points Rankings', includeLevels=False)
    
    def _doPointsRankings(self, title, includeEpisodes=True, includeLevels=True):
        self.checkCurHS()
        self.setText(title+'\n\n', 'h1')
        self.addLine("0th = 20 pts, 1st = 19 pts, ..., 19th = 1 pt\n")
        stats = getPoints(self.curHS, includeEpisodes, includeLevels).values()
        stats = [(x[1],x[0]) for x in stats if x[1]>0] #x iterates over players (keys)
        sortList(stats, reverse=True)
        for i in xrange(100):
            showIndex = i
            if i>0 and stats[i][0] == stats[i-1][0]:
                showIndex = '-'
            self.addLine('%2s  %s - %d' % (showIndex, self.dispPlayer(stats[i][1]), stats[i][0]))
      
    def avgPointsRankings(self):
        self.checkCurHS()
        self.setText('Average Points Rankings'+'\n\n', 'h1')
        stats = getPoints(self.curHS).values()
        stats = [[x[1],x[0]] for x in stats if x[1]>0] #x iterates over players (keys)
        dict = getAllPlayerStats(self.curHS)
        for i, (pts, player) in enumerate(stats):
            numberOf20 = dict[player.lower()][2]
            if numberOf20 > 19 :
                stats[i][0] /= float(numberOf20)
            else : stats[i][0] = 0
        sortList(stats, reverse=True)
        for i in xrange(100):
            showIndex = i
            if i>0 and stats[i][0] == stats[i-1][0]:
                showIndex = '-'
            if stats[i][0] == 0:
                break
            self.addLine('%2s  %s - %.3lf' % (showIndex, self.dispPlayer(stats[i][1]), stats[i][0]))
    
    def playerRankDis(self):
        #list of all non-top 20 scores for a given player
        self.checkCurHS()
        player = self.askPlayer('Player Rank Distribution',
            'Displays the number of times a player appears in each rank')
        if not player: return

        self.setText('Number of scores for each rank for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        rankDis = getRankDistribution(self.curHS, player)
        for i, e in enumerate(rankDis):
            self.addLine('%4s : %3d' % (getOrdinal(i), e))

    def _doTotalScoreReport(self, title):
        self.checkCurHS()
        self.setText(title+'\n\n', 'h1')
        total_lvl_score = getTotalScore(self.curHS)
        total_ep_score = getTotalScore(self.curHS, get_level=False)
        self.addLine("Community Total Level Score: %0.3f (%s)" % (total_lvl_score, getTime(total_lvl_score)))
        self.addLine("Community Total Episode Score: %0.3f (%s)" % (total_ep_score, getTime(total_ep_score)))
        self.addLine('')
        diff = total_lvl_score-total_ep_score-36000
        self.addLine("Difference between both : %0.3f (%s)" % (diff, getTime(diff)))
    
    def totalGameScore(self):
        self._doTotalScoreReport('Total Community Score')
    
    def _ZerothList(self, title, includeLevels=True, includeEpisodes=True):
        self.checkCurHS()
        self.setText("List of all "+title+'\n\n', 'h1')
        stats = getZerothList(self.curHS, includeLevels, includeEpisodes)
        for ep_num, lvl_num, entry in stats:
            self.addLine("%s - %s" % (self.dispLevel(ep_num, lvl_num), self.dispEntry(entry)))
            
    def _doTotalScoreRankings(self):
        self.checkCurHS()
        number = self.askNumber('How many players do you want in the rankings?', 'Players are taken from the top-20 rankings (min. 2)')
        if not number :
            return
        players = getNamesList(self.curHS, number)
        if not self.onlineScores:
            self.downloadOnlineScores(players)
        else :
            for k in self.onlineScores.keys():
                if k in players:
                    players.remove(k)
            if len(players) != 0:
                self.downloadOnlineScores(players)
        
        self.setText('Total Score Rankings\n\n', 'h1')
        stats = computeTotalScoreRankings(self.onlineScores).values()
        lvls = [(x[1],x[0]) for x in stats if x[1]>0]
        sortList(lvls, reverse=True)
        self.addLine('Total Level Score Rankings')
        for i in xrange(len(lvls)):
            shownIndex = i
            if i>0 and lvls[i][0] == lvls[i-1][0]:
                shownIndex = '-'
            self.addLine('%2s  %s - %.3f'%(shownIndex,
                self.dispPlayer(lvls[i][1]), lvls[i][0]))
        
        eps = [(x[2],x[0]) for x in stats if x[2]>0]
        sortList(eps, reverse=True)
        self.addLine('')
        self.addLine('')
        self.addLine('Total Episode Score Rankings')
        for i in xrange(len(eps)):
            shownIndex = i
            if i>0 and eps[i][0] == eps[i-1][0]:
                shownIndex = '-'
            self.addLine('%2s  %s - %.3f'%(shownIndex,
                self.dispPlayer(eps[i][1]), eps[i][0]))
        
        self.addLine('')
        self.addLine('Both rankings are made by taking only the first %d players of the top-20 rankings.' % number)
        self.addLine('Therefore, it can be inaccurate if a player only has a few top-20 but a good Total Score.')
    
        
    def playerTotalScore(self):
        player = self.askPlayer('Player Total Score',
            'Displays the total lvl and ep score for a player')
        if not player: return
        
        if not self.onlineScores:
            self.downloadOnlineScores([player])
        elif player not in self.onlineScores.keys():
            self.downloadOnlineScores([player])
        
        try :
            data = computePlayerTotalScore(self.onlineScores[player])
        except KeyError as e:
            raise NHighError("Player %s doesn't exist !" %str(e))
        
        self.setText('Total Score for %s:\n\n' %
                     self.dispPlayer(player),'h1')
        total_ep_score = data[1]
        total_lvl_score = data[0]
        self.addLine("Total Level Score: %0.3f (%s)" % (total_lvl_score, getTime(total_lvl_score)))
        self.addLine("Total Episode Score: %0.3f (%s)" % (total_ep_score, getTime(total_ep_score)))
        self.addLine('')
        diff = total_lvl_score-total_ep_score-36000
        self.addLine("Difference between both : %0.3f (%s)" % (diff, getTime(diff)))
    
    def playerMostImprovableEp(self):
        player = self.askPlayer("Player Most Improvable Episodes",
            'Displays the list of most improvable episodes for a player')
        if not player: return
        
        if not self.onlineScores:
            self.downloadOnlineScores([player])
        elif player not in self.onlineScores.keys():
            self.downloadOnlineScores([player])
        
        try:
            stats = computeEpLvlDiff(self.onlineScores[player])
        except KeyError as e:
            raise NHighError("Player %s doesn't exist !" %str(e))
        self.setText('List of most improvable episode scores for %s:\n\n' %
                self.dispPlayer(player),'h1')
        for i, (ep, epScore, diff) in enumerate(stats):
            self.addLine('%2s - %s - Current : %s Diff: %s' % (i, self.dispLevel(ep, 5),
                    self.dispTime(epScore), self.dispTime(diff)))
        
        self.addLine('')
        self.addLine('The difference is the difference between actual episode score and the sum of the five level scores')
    
    def playerPossibleEp0th(self):
        self.checkCurHS()
        player = self.askPlayer("Player Possible Episode 0th",
            'Displays the list of episode 0th a player can easily take')
        if not player: return
        
        if not self.onlineScores:
            self.downloadOnlineScores([player])
        elif player not in self.onlineScores.keys():
            self.downloadOnlineScores([player])
            
        try:    
            stats = findPossibleEp0th(self.curHS, self.onlineScores[player], player)
        except KeyError as e:
            raise NHighError("Player %s doesn't exist !" %str(e))
        if len(stats) == 0:
            self.setText('No possible episode 0th found for %s' % self.dispPlayer(player))
        else :
            self.setText('List of possible episode 0th for %s:\n\n' %
                         self.dispPlayer(player),'h1')
            
            for i,(ep, ep0th, best, diff) in enumerate(stats):
                self.addLine('%2s - %s - 0th : %s - %s: %s Diff: %s' % (i, self.dispLevel(ep, 5),
                        self.dispEntry(ep0th), self.dispPlayer(player), self.dispTime(best) ,self.dispTime(diff)))
            
            self.addLine('')
            self.addLine("The difference is the difference between actual episode 0th and the sum of the player's five level scores")
    
    def playerNon20List_wDiff(self):
        #list of all non-top 20 scores for a given player
        self.checkCurHS()
        player = self.askPlayer('Player Non-Top-20 List',
            'Displays a list of all non-top-20 scores held by a player')
        if not player: return
        
        if not self.onlineScores:
            self.downloadOnlineScores([player])
        elif player not in self.onlineScores.keys():
            self.downloadOnlineScores([player])
        
        try :
            nontwenty = computePlayerNonScores(self.curHS, self.onlineScores[player], player)
        except KeyError as e:
            raise NHighError("Player %s doesn't exist !" %str(e))
        if len(nontwenty) == 0:
            self.setText('No non-top-20 score for %s' % player)
        else : 
            self.setText('List of non-top-20 scores for %s:\n\n' %
                         self.dispPlayer(player), 'h1')
            for (ep_num, lvl_num, diff) in nontwenty:
                self.addLine('%s - Diff : %s' %
                    (self.dispLevel(ep_num, lvl_num), self.dispTime(diff)))
            self.addLine('\nTotal: %d non-top-20 scores' % len(nontwenty))
    
    def playerAllWorstLvlList(self):
        self.checkCurHS()
        player = self.askPlayer('Player Worst Level Scores List',
            'Displays a list of the worst scores held by a player (in terms of diff to 0th)')
        if not player : return
        
        if not self.onlineScores:
            self.downloadOnlineScores([player])
        elif player not in self.onlineScores.keys():
            self.downloadOnlineScores([player])
        
        try :
            list = computePlayerWorstLvlScores(self.curHS, self.onlineScores[player], player)
        except KeyError as e:
            raise NHighError("Player %s doesn't exist !" %str(e))
        
        self.setText('List of worst level top-20 scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1') 
        for i, (ep, lvl, entry, diff) in enumerate(list[:100]) :
            self.addLine('%2s - %s - %s Diff: %s' %
                (i, self.dispLevel(ep, lvl), 
                self.dispTime(entry), self.dispTime(diff)))
#end of edit

    def levelTop20(self):
        self.checkCurHS()
        ep, lvl = self.askLevelOrEpisode('Level/Episode Top 20',
            'Shows the top-20 for any level or episode (as shown in the game)')
        if ep is None:
            return
        if lvl==5:
            title = 'Top-20 For Episode <level>%d</level>'%ep
        else:
            title = 'Top-20 For Level <level>%d-%d</level>'%(ep,lvl)
        
        self.setText(title+'\n\n', tags='h1')

        for pos,entry in enumerate(self.curHS.table[ep][lvl]):
            if entry.score:
                self.addLine('%2d. %s' % (pos, self.dispEntry(entry)))

    def demoAnalysis(self):
        SEL_TOP20, SEL_DEMO = 'top20', 'demo'
        frames = [
            AskDlgSelectionSubFrame(
                [
                    AskDlgHorizSubFrame([
                        IntField('episode','Episode',minValue=0,maxValue=99, width=3),
                        IntField('level','Level',minValue=0,maxValue=4, width=3),
                        IntField('rank','Rank',minValue=0,maxValue=19, width=3),
                    ]),
                    MultiLineField('demo','Demo data'),
                ],
                ['Online Top-20 Score', 'Demo Data'],
                [SEL_TOP20, SEL_DEMO]),
        ]
        dlg = AskDialog(self.root, 'Demo Analysis',
            'Description: Analyzes a demo and shows\nwhich keys were pressed for each frame',
            frames)
        res = dlg.run()
        if not res:
            return
        if res.selection==SEL_TOP20:
            player, score, demo = downloadReplay(res.episode, res.level, res.rank)
            title = 'Demo analysis for %s (%s rank) - [%s, %s]' % \
                    (self.dispLevel(res.episode, res.level),
                     getOrdinal(res.rank), self.dispPlayer(player),
                     self.dispTime(score*FRAME_TIME))
        elif res.selection==SEL_DEMO:
            demo = res.demo
            title = 'Demo analysis'
        else:
            return

        arr = parseAndCompactReplay(demo)
        self.setText(title+'\n\n')
        for frame, key, num in arr:
            self.addLine('%03d. %-7s - %d frame%s' %
                (frame, getReplayKeyName(key), num, getPlural(num)))

    def unsubmitted20Report(self):
        self.checkCurHS()
        solData = readSolFile()
        player, res = findUnsubmittedTop20(solData, self.curHS)
        self.setText('Unsubmitted top-20 scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        if not res:
            self.addLine('No unsubmitted scores found')
        else:
            for ep, lvl, newRank, newScore, oldRank, oldEntry in res:
                newTime = FRAME_TIME * newScore
                oldEntryTxt = ''
                if oldEntry:
                    oldEntryTxt = 'Improved from %4s - %s' % \
                        (getOrdinal(oldRank), self.dispEntry(oldEntry))
                self.addLine('%s - %4s - %s  %s' % \
                    (self.dispLevel(ep,lvl), getOrdinal(newRank),
                     self.dispTime(newTime), oldEntryTxt))
        
    def totalScoreReport(self):
        def getTimeDisp(score):
            hours, frames = divmod(score, (FRAMES_PER_SEC*60*60))
            mins, frames = divmod(frames, (FRAMES_PER_SEC*60))
            secs = frames * FRAME_TIME
            return '%d:%02d:%06.3f' % (hours,mins,secs)
        solData = readSolFile()
        lvlScore, epScore = calcTotalScores(solData)
        self.setText('Game total scores:\n\n','h1')
        self.addLine('Total level score: %0.3f (%s)' % (lvlScore*FRAME_TIME, getTimeDisp(lvlScore)))
        self.addLine('Total episode score: %0.3f (%s)' % (epScore*FRAME_TIME, getTimeDisp(epScore)))

#eru_bahagon added 2 following function
    def nonTop20Scores(self):
        self.checkCurHS()
        solData = readSolFile()
        player, res = findNonTop20Scores(solData, self.curHS)
        self.setText('Non-Top-20 scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        if not res:
            self.addLine('Congrats, You have all possible highscores')
        else :
            for ep, lvl, diff in res:
                self.addLine('%s - Diff : %s' % (self.dispLevel(ep, lvl), self.dispTime(diff*FRAME_TIME)))
            self.addLine('\nTotal: %d non-top-20 scores' % len(res))
    
    def persMostImprovableEp(self):
        self.checkCurHS()
        solData = readSolFile()
        player, res = findEpLvlDifference(solData, self.curHS)
        self.setText('Most improvable episode scores for %s:\n\n' %
                     self.dispPlayer(player), 'h1')
        self.addLine("Please, make sure you've submitted all scores\n\n")
        for i, (ep, ep_entry, diff) in enumerate(res) :
            self.addLine('%2s - %s - %s Diff: %s' % (i, self.dispLevel(ep, 5),
                    self.dispEntry(ep_entry), self.dispTime(diff)))
#end of edit        

    def dispLevel(self, ep, level):
        return '<level>%-4s</level>' % getLevelName(ep, level)
    def dispPlayer(self, player):
        return '<player>' + player + '</player>'
    def _dispPlayer(self, player):
        return '<player>%14s</player>' % player
    def dispTime(self, time):
        return '<time>%7.3f</time>' % time
    def dispEntry(self, entry):
        if COLUMN_REP == 1:
            return '[%s, %s]'%(self._dispPlayer(entry.name),
                           self.dispTime(entry.time))
        else :
            return '[%s, %s]'%(self.dispPlayer(entry.name),
                           self.dispTime(entry.time))

    re_tag = re.compile('<([^>]+)>')
    def _addTaggedText(self, text, tags=None):
        if tags:
            tagList = tags.split()
        else:
            tagList = []
        while text:
            m = self.re_tag.search(text)
            if m:
                textNow = text[:m.start()]
                text = text[m.end():]
            else:
                textNow = text
                text = ''
                
            if textNow:
                self.textWidget.insert(END, textNow, ' '.join(tagList))

            if m:
                tag = m.group(1)
                if tag.startswith('/'):
                    #remove last tag
                    tagList.pop()
                else:
                    tagList.append(tag)

    def setText(self, text, tags=None):
        self.textWidget.config(state=NORMAL)
        self.textWidget.delete('1.0', END)
        self._addTaggedText(text, tags)
        self.textWidget.config(state=DISABLED)

    def addText(self, txt, tags=None):
        self.textWidget.config(state=NORMAL)
        self._addTaggedText(txt, tags)
        self.textWidget.config(state=DISABLED)

    def clearText(self):
        self.setText('')

    def addLine(self, txt, tags=None):
        self.addText(txt + '\n', tags)

    def options(self):
        dlg = OptionsDlg(self.root)
        dlg.run()

    def findIcon(self):
        if sys.platform.startswith("win"):
            if sys.argv[0].lower().endswith('.exe'):
                return sys.argv[0]
            else:
                here = os.path.dirname(sys.argv[0])
                ico = os.path.join(here, 'nhighico.ico')
                if os.path.exists(ico):
                    return ico
        return None
    
    def selectAllText(self, event):
        'Control-A handler on Text class'
        event.widget.tag_add(SEL, '1.0', END)

    def _getClickedTagContent(self, tagName):
        curIdx = self.textWidget.index('current')
        prevList = self.textWidget.tag_prevrange(tagName,
            'current + 1 chars', 'current linestart')
        if not prevList:
            return None
        tagStart, tagStop = prevList
        tagStartInts = map(int, tagStart.split('.'))
        tagStopInts = map(int, tagStop.split('.'))
        curIdxInts = map(int, curIdx.split('.'))
        if tagStartInts > curIdxInts or tagStopInts <= curIdxInts:
            return None
        return self.textWidget.get(tagStart, tagStop)

    def showLevelContextMenu(self, event):
        levelText = self._getClickedTagContent('level')
        if not levelText:
            return

        m = re.match(r'(\d+)-(\d)', levelText)
        if m:
            ep, lvl = map(int, m.groups())
            typ = 'Level'
        elif re.match(r'(\d+)', levelText):
            ep, lvl = int(levelText), 5
            typ = 'Episode'

        self.lastRightClickedLevel = (ep,lvl)
        self.levelContextMenu.entryconfig(0,
            label='Reports for %s %s:' % (typ, levelText))
        self.levelContextMenu.tk_popup(event.x_root, event.y_root)
        
    def showPlayerContextMenu(self, event):
        player = self._getClickedTagContent('player')
        if not player:
            return

        self.lastRightClickedPlayer = player
        self.playerContextMenu.entryconfig(0, label='Reports for '+player+':')
        self.playerContextMenu.tk_popup(event.x_root, event.y_root)

    def show_hand_cursor(self, event):
        event.widget.configure(cursor="hand2")
    def show_arrow_cursor(self, event):
        event.widget.configure(cursor="xterm")

    def buildRootWin(self, win):
        win.title(WIN_TITLE)
        self.iconfile = self.findIcon()
        
        curHSFrame = Frame(win)
        lbl1 = Label(curHSFrame, text="Current highscores from:")
        lbl1.grid(row=0,column=0)
        self.txtCurHSDate = StringVar()
        self.txtCurHSDate.set('<Not Loaded>')
        lbl2 = Label(curHSFrame, textvariable=self.txtCurHSDate)
        lbl2.grid(row=0,column=1, sticky=W)
        curHSFrame.columnconfigure(1, weight=1)
        curHSFrame.grid(row=0, column=0)

        textFrame = Frame(win)
        self.textWidget = Text(textFrame, wrap=NONE)
        textVScroll = Scrollbar(textFrame, command=self.textWidget.yview)
        textVScroll.grid(row=0, column=1, sticky=N+S)
        textHScroll = Scrollbar(textFrame, orient=HORIZONTAL,
                                command=self.textWidget.xview)
        textHScroll.grid(row=1, column=0, sticky=E+W)
        self.textWidget.config(yscrollcommand=textVScroll.set,
                               xscrollcommand=textHScroll.set)

        ##### colors/fonts #####
        self.textWidget.config(foreground='#000000', background='#FFFFFF', font='Courier 12')
        self.textWidget.tag_config('h1', font='Courier 12 bold')
        self.textWidget.tag_config('h2', font='Courier 12 bold')
        self.textWidget.tag_config('player', foreground='#800000')
        self.textWidget.tag_config('time', foreground='#404080')
        self.textWidget.tag_config('level', foreground='#008000')

        ##### context menus #####
        self.textWidget.tag_bind('player', '<ButtonRelease-1>', self.showPlayerContextMenu)
        self.textWidget.tag_bind('player', '<ButtonRelease-3>', self.showPlayerContextMenu)
        self.textWidget.tag_bind('player', '<Enter>', self.show_hand_cursor)
        self.textWidget.tag_bind('player', '<Leave>', self.show_arrow_cursor)
        self.textWidget.tag_bind('level', '<ButtonRelease-1>', self.showLevelContextMenu)
        self.textWidget.tag_bind('level', '<ButtonRelease-3>', self.showLevelContextMenu)
        self.textWidget.tag_bind('level', '<Enter>', self.show_hand_cursor)
        self.textWidget.tag_bind('level', '<Leave>', self.show_arrow_cursor)
        
        self.textWidget.bind_class('Text', '<Control-Key-A>', self.selectAllText)
        self.textWidget.bind_class('Text', '<Control-Key-a>', self.selectAllText)
        
        self.addText(welcomeStr)
        self.textWidget.grid(row=0, column=0, sticky=N+E+W+S)
        textFrame.columnconfigure(0, weight=1)
        textFrame.rowconfigure(0, weight=1)
        textFrame.grid(row=1, column=0, sticky=N+E+W+S)
        win.columnconfigure(0, weight=1)
        win.rowconfigure(1, weight=1)

        self.defineMenuItems()                               
        self.buildMenus(win)
        win.update_idletasks()
        if self.iconfile:
            win.iconbitmap(self.iconfile)

    def addKeyBinding(self, shortName, handler):
        def dummyKeyPress(event):
            pass
        
        def addKeyBindingHelper(pattern):
            self.root.bind(pattern, handler)
            
            #make sure we don't get called if additional modifiers are pressed
            for modifier in ['Control-', 'Shift-', 'Alt-']:
                if modifier.lower() not in pattern.lower():
                    self.root.bind('<' + modifier + pattern[1:], dummyKeyPress, '+')
        
        #convert short name (Ctrl+O) to Tk event pattern (<Control-Key-O>)
        s = shortName
        s = s.replace('+', '-')
        s = re.sub('(?i)Ctrl','Control', s)
        idx = s.rfind('-') + 1 #position after last '-', or start of string
        pattern = '<' + s[:idx] + 'Key-' + s[idx:] + '>'

        addKeyBindingHelper(pattern)
        if re.search('-[A-Za-z]>$', pattern):
            pattern = pattern[:-2] + pattern[-2].swapcase() + pattern[-1:]
            addKeyBindingHelper(pattern)            
        
    def buildMenu(self, win, menuItems):
        def getLabelAndShortcut(label):
            label = label.replace('&&', '\x01') #just a temp value
            underline = label.find('&')
            if underline<0:
                underline = None
            label = label.replace('&','')
            label = label.replace('\x01','&')
            return label, underline
        
        def do_submenu(parent, items):
            menu = Menu(parent, tearoff=0)
            for item in items:
                if type(item) is MenuSeparator:
                    menu.add_separator()
                else:
                    label, underline = getLabelAndShortcut(item.label)
                    if hasattr(item,'command') and item.command:
                        accel = None
                        if hasattr(item,'key') and item.key:
                            accel = item.key
                            self.addKeyBinding(item.key, item.wrappedCommand)
                        menu.add_command(label=label, command=item.wrappedCommand,
                                         underline=underline, accelerator=accel)
                    elif hasattr(item,'items') and item.items:
                        submenu = do_submenu(menu, item.items)
                        menu.add_cascade(label=label, menu=submenu,
                                         underline=underline)
                    else:
                        menu.add_command(label=label, state=DISABLED)
            return menu
        return do_submenu(win, menuItems)
        
    def buildMenus(self, win):
        menu = self.buildMenu(win, self.menuItems)
        win.config(menu=menu)

        self.playerContextMenu = self.buildMenu(win, self.playerContextMenuItems)
        self.levelContextMenu = self.buildMenu(win, self.levelContextMenuItems)

    def quitProgram(self):
        self.root.quit()

    def defineMenuItems(self):

        ###### Context menu for a level or episode #######

        self.levelContextMenuItems = [
            MenuItem("Reports for xxxxxx", None),
            MenuSeparator(),
            ContextMenuItem("Top-20", self.levelTop20),
        ]

        ###### Context menu for a player #######

        self.playerContextMenuItems = [
            MenuItem("Reports for xxxxxx", None),
            MenuSeparator(),
            ContextMenuItem("0th Changes", self.player0thChangesReport),
            ContextMenuItem("Top-20 Changes", self.player20ChangesReport),
            ContextMenuItem("Statistics", self.playerStatisticsReport),
            MenuSeparator(),
            ContextMenuItem("0th Scores", self.player0thList),
            ContextMenuItem("Top-20 Scores", self.player20List),
            ContextMenuItem("Episode Top-20 Scores", self.playerEp20List),
            ContextMenuItem("Top-X Scores", self.playerXYList),
            ContextMenuItem("Rank Distribution", self.playerRankDis),
            MenuSeparator(),
            ContextMenuItem("Non-Top-20 Scores", self.playerNon20List),
            ContextMenuItem("Worst Level Scores", self.playerWorstLvlList),
            ContextMenuItem("Worst Episode Scores", self.playerWorstEpList),
            ContextMenuItem("Comparison with ...", self.playerComparison),
            MenuSeparator(),
            ContextMenuItem("Total Score", self.playerTotalScore),
            ContextMenuItem("Most Improvable Episodes", self.playerMostImprovableEp),
            ContextMenuItem("Possible Episode 0ths", self.playerPossibleEp0th),
            MenuSeparator(),
            ContextMenuItem("Add to Ignore List", self.addPlayerToIgnoreList),
        ]

        ###### Main Menu  ######        
        self.menuItems = [
            SubMenu('File',
            [
                MenuItem('&Download', self.downloadHS, key='Ctrl+D'),
                MenuItem('&Load', self.loadHS, key='Ctrl+O'),
                MenuItem('&Save As...', self.saveHS, key='Ctrl+S'),
                MenuSeparator(),
                MenuItem('E&xit', self.quitProgram),
            ]),
            SubMenu('Reports',
            [
                MenuItem('New 0th scores', self.newZerothReport, key='F1'),
                MenuSeparator(),
                MenuItem("0th Rankings", self.mostZerothReport, key='F2'),
                MenuItem("Top-20 Rankings", self.mostScoresReport, key='F3'),
                MenuItem("Top-X Rankings", self.mostXYReport),
                SubMenu("Levels...",
                [
                    MenuItem("Level 0th Rankings", self.mostZerothReportNoEpisodes),
                    MenuItem("Level Top-20 Rankings", self.mostLvlScoresReport),
                    MenuSeparator(),
                    MenuItem("Largest Spread Levels (0th-1st)", lambda: self.difficultLevelsReport(0,1)),
                    MenuItem("Largest Spread Levels (0th-19th)", lambda: self.difficultLevelsReport(0,19)),
                    MenuSeparator(),
                    MenuItem("Smallest Spread Levels (0th-1st)", lambda: self.easyLevelsReport(0,1)),
					MenuItem("Smallest Spread Levels (0th-w/o tie)", lambda : self.easyLevelsReport(0,1, incl_tie=False)),
                    MenuItem("Smallest Spread Levels (0th-19th)", lambda: self.easyLevelsReport(0,19)),
                ]),
				SubMenu("Episodes...", 
				[
					MenuItem("Episode 0th Rankings", self.mostZerothReportNoLevels),
					MenuItem("Episode Top-20 Rankings", self.mostEpScoresReport),
                    MenuItem("Episode Owned", self.ownedEp),
                    MenuSeparator(),
					MenuItem("Most improvable episodes", self.mostImprovableEps),
					MenuItem("Least improvable episodes", self.leastImprovableEps),
                    MenuSeparator(),
                    MenuItem("Largest Spread Episodes (0th-1st)", lambda: self.difficultEpsReport(0,1)),
                    MenuItem("Largest Spread Episodes (0th-19th)", lambda: self.difficultEpsReport(0,19)),
                    MenuSeparator(),
                    MenuItem("Smallest Spread Episodes (0th-1st)", lambda: self.easyEpsReport(0,1)),
                    MenuItem("Smallest Spread Episodes (0th-19th)", lambda: self.easyEpsReport(0,19)),
				]),
                SubMenu("Points Rankings...",
                [
                    MenuItem("Total Points Rankings", self.mostPoints, key='F6'),
                    MenuItem("Level Points Rankings", self.mostLvlPoints),
                    MenuItem("Episode Points Rankings", self.mostEpPoints),
                    MenuItem("Average Points Rankings", self.avgPointsRankings)
                ]),
                MenuSeparator(),
				SubMenu("Player's Statistics...", 
				[
					MenuItem("Player's 0th Changes", self.player0thChangesReport, key='F4'),
					MenuItem("Player's top-20 Changes", self.player20ChangesReport, key='F5'),
					MenuItem("Player's statistics", self.playerStatisticsReport),
                    MenuSeparator(),
					MenuItem("Player's 0th scores", self.player0thList),
					MenuItem("Player's top-20 scores", self.player20List),
                    MenuItem("Player's ep top-20 scores", self.playerEp20List),
                    MenuItem("Player's top-X scores", self.playerXYList),
                    MenuItem("Player's rank distribution", self.playerRankDis),
                    MenuSeparator(),
                    MenuItem("Player's non-top-20 scores", self.playerNon20List),
                    MenuItem("Player's worst level scores", self.playerWorstLvlList),
                    MenuItem("Player's worst episode scores", self.playerWorstEpList),
                    MenuItem("Player's comparison with ...", self.playerComparison),
                    #MenuItem("Player's possible episode 0ths", self.playerPossibleEp0th),
				]),
                MenuSeparator(),
                MenuItem("Community Total Score", self.totalGameScore),
                MenuSeparator(),
                MenuItem("Level/Episode top-20", self.levelTop20),
                SubMenu("0th List...",
                [
                    MenuItem("Complete 0th List", lambda : self._ZerothList("0th", True, True)),
                    MenuItem("Level 0th List", lambda : self._ZerothList("level 0th", True, False)),
                    MenuItem("Episode 0th List", lambda : self._ZerothList("episode 0th", False, True))
                ])
            ]),
            SubMenu('Personal Data (.sol)',
            [
                MenuItem('Find unsubmitted top-20', self.unsubmitted20Report),
                MenuItem('Total game score', self.totalScoreReport),
                MenuItem('Find non-top-20 scores', self.nonTop20Scores),
                MenuItem('Find most improvable episodes', self.persMostImprovableEp),
            ]),
            SubMenu('Online Scores',
            [
                MenuItem("Total Score Rankings", self._doTotalScoreRankings),
                MenuSeparator(),
                MenuItem("Player's total score", self.playerTotalScore),
                MenuItem("Player's most improvable episodes", self.playerMostImprovableEp),
                MenuItem("Player's possible episode 0ths", self.playerPossibleEp0th),
                MenuItem("Player's non-top-20 scores", self.playerNon20List_wDiff),
                MenuItem("Player's worst level scores", self.playerAllWorstLvlList),
            ]),
            SubMenu('Tools',
            [
                MenuItem('Options...', self.options),
                MenuItem('Demo Analysis...', self.demoAnalysis),
            ]),
        ]


app = None
def run():
    root = Tk()
    global app
    app = NHighApp(root)
    app.build()
    app.autoLoadHS()
    root.mainloop()

def safeRun():
    try:
        run()
    except KeyboardInterrupt:
        pass
    except (StandardError, TclError):
        print >>sys.stderr, '**** Exception occurred ****'
        print >>sys.stderr
        import traceback
        traceback.print_exc()

if __name__=='__main__':
    safeRun()
